@@ -7,7 +7,7 @@
 # Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
 #
 
-from mo_dots import Data, Null, coalesce, is_data, is_list, to_data
+from mo_dots import Data, Null, coalesce, is_data, is_list, to_data, is_many, unwraplist
 from mo_future import PY2, is_text, text, unichr, urlparse, is_binary
 from mo_logs import Log
 
@@ -57,12 +57,26 @@ class URL(object):
             Log.error(u"problem parsing {{value}} to URL", value=value, cause=e)
 
     def __nonzero__(self):
-        if self.scheme or self.host or self.port or self.path or self.query or self.fragment:
+        if (
+            self.scheme
+            or self.host
+            or self.port
+            or self.path
+            or self.query
+            or self.fragment
+        ):
             return True
         return False
 
     def __bool__(self):
-        if self.scheme or self.host or self.port or self.path or self.query or self.fragment:
+        if (
+            self.scheme
+            or self.host
+            or self.port
+            or self.path
+            or self.query
+            or self.fragment
+        ):
             return True
         return False
 
@@ -70,7 +84,7 @@ class URL(object):
         if not is_text(other):
             Log.error(u"Expecting text path")
         output = self.__copy__()
-        output.path = output.path.rstrip('/') + "/" + other.lstrip('/')
+        output.path = output.path.rstrip("/") + "/" + other.lstrip("/")
         return output
 
     def __add__(self, other):
@@ -81,7 +95,7 @@ class URL(object):
         return output
 
     def __unicode__(self):
-        return self.__str__().decode('utf8')  # ASSUME chr<128 ARE VALID UNICODE
+        return self.__str__().decode("utf8")  # ASSUME chr<128 ARE VALID UNICODE
 
     def __copy__(self):
         output = URL(None)
@@ -93,7 +107,7 @@ class URL(object):
         output.fragment = self.fragment
         return output
 
-    def decode(self, encoding=''):
+    def decode(self, encoding=""):
         return text(self).decode(encoding)
 
     def __data__(self):
@@ -110,7 +124,7 @@ class URL(object):
         if self.host:
             url = self.host
         if self.scheme:
-            url = self.scheme + "://"+url
+            url = self.scheme + "://" + url
         if self.port:
             url = url + ":" + str(self.port)
         if self.path:
@@ -135,18 +149,21 @@ def hex2chr(hex):
     except Exception as e:
         raise e
 
+
 if PY2:
     _map2url = {chr(i): chr(i) for i in range(32, 128)}
-    for c in " {}<>;/?:@&=+$,":
+    for c in "{}<>;/?:@&=+$%,+":
         _map2url[c] = "%" + str(int2hex(ord(c), 2))
     for i in range(128, 256):
         _map2url[chr(i)] = "%" + str(int2hex(i, 2))
+    _map2url[chr(32)] = "+"
 else:
     _map2url = {i: unichr(i) for i in range(32, 128)}
-    for c in b" {}<>;/?:@&=+$,":
+    for c in b"{}<>;/?:@&=+$%,+":
         _map2url[c] = "%" + int2hex(c, 2)
     for i in range(128, 256):
         _map2url[i] = "%" + str(int2hex(i, 2))
+    _map2url[32] = "+"
 
 
 names = ["path", "query", "fragment"]
@@ -163,7 +180,11 @@ def parse(output, suffix, curr, next):
         parse(output, suffix, curr, next + 1)
     else:
         output.__setattr__(names[curr], suffix[:e:])
-        parse(output, suffix[e + 1::], next, next + 1)
+        parse(output, suffix[e + 1 : :], next, next + 1)
+
+
+def hex2byte(v):
+    return bytes([int(v, 16)])
 
 
 def url_param2value(param):
@@ -175,30 +196,99 @@ def url_param2value(param):
     if param == None:
         return Null
 
-    def _decode(v):
-        output = []
-        i = 0
-        while i < len(v):
-            c = v[i]
-            if c == "%":
-                d = hex2chr(v[i + 1:i + 3])
-                output.append(d)
-                i += 3
-            else:
-                output.append(c)
-                i += 1
+    def _decode(vs):
+        from mo_json import json2value
 
-        output = text("".join(output))
-        try:
-            from mo_json import json2value
+        results = []
+        for v in vs.split(","):
+            output = []
+            i = 0
+            utf_remaining = 0
+            start = 0
+            while i < len(v):
+                c = v[i]
+                if utf_remaining:
+                    if c == "%":
+                        try:
+                            hex = v[i + 1 : i + 3]
+                            if hex.strip() == hex:
+                                d = int(v[i + 1: i + 3], 16)
+                                if d & 0xC0 == 0x80:  # 10XX xxxx
+                                    utf_remaining -= 1
+                                    b = bytes([d])
+                                    output.append(b)
+                                    i += 3
+                                    continue
+                        except Exception:
+                            pass
+                    # missing continuation byte, try again
+                    output = output[:-utf_remaining]
+                    utf_remaining = 0
+                    i = start
+                    output.append(b"%")
+                    i += 1
+                else:
+                    if c == "+":
+                        output.append(b" ")
+                        i += 1
+                    elif c == "%":
+                        try:
+                            hex = v[i + 1 : i + 3]
+                            if hex.strip() != hex:
+                                output.append(b"%")
+                                i += 1
+                                continue
 
-            return json2value(output)
-        except Exception:
-            pass
-        return output
+                            d = int(v[i + 1: i + 3], 16)
+                            if d & 0xE0 == 0xC0:  # 110X xxxx
+                                utf_remaining = 1
+                                start = i
+                                b = bytes([d])
+                                output.append(b)
+                                i += 3
+                            elif d & 0xF0 == 0xE0:  # 1110 xxxx
+                                utf_remaining = 2
+                                start = i
+                                b = bytes([d])
+                                output.append(b)
+                                i += 3
+                            elif d & 0xF8 == 0xF0:  # 1111 0xxx
+                                utf_remaining = 3
+                                start = i
+                                b = bytes([d])
+                                output.append(b)
+                                i += 3
+                            elif d & 0x80 == 0x80:
+                                output.append(b"%")
+                                i += 1
+                            else:
+                                b = bytes([d])
+                                output.append(b)
+                                i += 3
+                        except Exception:
+                            output.append(b"%")
+                            i += 1
+                    else:
+                        try:
+                            output.append(c.encode("latin1"))
+                        except Exception:
+                            output.append(c.encode('utf8'))
+                        i += 1
+
+            if utf_remaining:
+                # missing continuation byte, try again
+                output = output[:-utf_remaining] + [v[start:].encode("latin1")]
+
+            output = b"".join(output).decode("utf8")
+            try:
+                output = json2value(output)
+            except Exception:
+                pass
+            results.append(output)
+        return unwraplist(results)
 
     query = Data()
-    for p in param.split('&'):
+    for p in param.split("&"):
         if not p:
             continue
         if p.find("=") == -1:
@@ -206,6 +296,7 @@ def url_param2value(param):
             v = True
         else:
             k, v = p.split("=")
+            k = _decode(k)
             v = _decode(v)
 
         u = query.get(k)
@@ -224,26 +315,34 @@ def value2url_param(value):
     :param value:
     :return: ascii URL
     """
+    from mo_json import value2json, json2value
+
+    def _encode(value):
+        return "".join(_map2url[c] for c in value.encode("utf8"))
+
     if value == None:
-        Log.error("Can not encode None into a URL")
+        return None
 
     if is_data(value):
-        from mo_json import value2json
-
         value_ = to_data(value)
         output = "&".join(
-            value2url_param(k) + "=" + (value2url_param(v) if is_text(v) else value2url_param(value2json(v)))
+            kk + "=" + vv
             for k, v in sorted(value_.leaves(), key=lambda p: p[0])
+            for kk, vv in [(value2url_param(k), value2url_param(v))]
+            if vv or vv == 0
         )
     elif is_text(value):
-        output = "".join(_map2url[c] for c in value.encode('utf8'))
+        try:
+            json2value(value)
+            output = _encode(value2json(value))
+        except Exception:
+            output = _encode(value)
     elif is_binary(value):
         output = "".join(_map2url[c] for c in value)
-    elif hasattr(value, "__iter__"):
-        output = ",".join(value2url_param(v) for v in value)
+    elif is_many(value):
+        output = ",".join(
+            vv for v in value for vv in [value2url_param(v)] if vv or vv == 0
+        )
     else:
-        output = str(value)
+        output = _encode(value2json(value))
     return output
-
-
-
